ASYNC_PATH = "async"
ASYNC_JOB_RESULT_PATH = "/{request_id}/results"
ASYNC_JOB_STATUS_PATH = "/{request_id}/status"
AUTH_URL = "https://{subdomain}.auth.marketingcloudapis.com/"
CUSTOM_OBJECT_DATA_PATH = "/customobjectdata/key/{key}/rowset"
DATA_EXTENSION_KEY_PATH = "/dataextensions/key:{key}/rows"
DATA_EVENT_KEY_PATH = "/dataevents/key:{key}/rowset"
DATA_PATH = "data"
FILTER_QUERY = "$filter={filter}"
FORWARD_SLASH = "/"
GRANT_TYPE = "client_credentials"
QUERY_PATH = "/services/data/v64.0/query?q="
SERVICE_CLOUD_AUTH_URL = "https://{subdomain}.salesforce.com/services/oauth2"
SERVICE_CLOUD_GRANT_TYPE = "refresh_token"
SYNC_PATH = "hub"
TOKEN_PATH = "token"
UPDATE_ACCOUNT_PATH = "/services/apexrest/account/update"
UPSERT_ACCOUNT_PATH = "/services/apexrest/account/upsert"
V1 = "v1"
V2 = "v2"
